import os
import shutil
from textwrap import dedent

from collections import defaultdict
from gi.repository import GLib, Gio

SYSTEM_BUS = Gio.bus_get_sync(Gio.BusType.SYSTEM)


def get_unit_names(region, arch, ns):
    if arch == 'amd64':
        unit_names = ['autopkgtest@{}-{}.service'.format(region, n) for n in ns]
    else:
        unit_names = ['autopkgtest@{}-{}-{}.service'.format(region, arch, n) for n in ns]

    return unit_names


def reload():
    SYSTEM_BUS.call_sync('org.freedesktop.systemd1',
                         '/org/freedesktop/systemd1',
                         'org.freedesktop.systemd1.Manager',
                         'Reload',
                         None,  # parameters
                         None,  # reply type
                         Gio.DBusCallFlags.NONE,
                         -1,  # timeout
                         None)  # cancellable


def enabledisable(unit_names, enabledisable, enabledisablearg, startstop):
    print('calling {enabledisable} then {startstop} on {unit_names}'.format(**locals()))
    SYSTEM_BUS.call_sync('org.freedesktop.systemd1',
                         '/org/freedesktop/systemd1',
                         'org.freedesktop.systemd1.Manager',
                         enabledisable,
                         enabledisablearg,
                         GLib.VariantType('*'),  # reply type, using '*' because we don't actually care
                         Gio.DBusCallFlags.NONE,
                         -1,  # timeout
                         None)  # cancellable

    for unit in unit_names:
        (_, ) = SYSTEM_BUS.call_sync('org.freedesktop.systemd1',
                                     '/org/freedesktop/systemd1',
                                     'org.freedesktop.systemd1.Manager',
                                     startstop,
                                     GLib.Variant('(ss)', (unit, 'replace')),
                                     GLib.VariantType('(o)'),  # reply type
                                     Gio.DBusCallFlags.NONE,
                                     -1,  # timeout
                                     None)  # cancellable


def enable(unit_names):
    enabledisable(unit_names,
                  'EnableUnitFiles',
                  GLib.Variant('(asbb)', (unit_names, False, True)),
                  'StartUnit')
    reload()


def disable(unit_names):
    enabledisable(unit_names,
                  'DisableUnitFiles',
                  GLib.Variant('(asb)', (unit_names, False)),
                  'StopUnit')
    for unit in unit_names:
        SYSTEM_BUS.call_sync('org.freedesktop.systemd1',
                             '/org/freedesktop/systemd1',
                             'org.freedesktop.systemd1.Manager',
                             'ResetFailedUnit',
                             GLib.Variant('(s)', (unit, )),
                             None,  # reply type
                             Gio.DBusCallFlags.NONE,
                             -1.,  # timeout
                             None)  # cancellable
        try:
            dropindir = os.path.join(os.path.sep,
                                     'etc',
                                     'systemd',
                                     'system',
                                     '{}.d'.format(unit))
            shutil.rmtree(dropindir)
        except FileNotFoundError:
            pass

        reload()


def get_units():
    (units, ) = SYSTEM_BUS.call_sync('org.freedesktop.systemd1',
                                     '/org/freedesktop/systemd1',
                                     'org.freedesktop.systemd1.Manager',
                                     'ListUnits',
                                     None,  # parameters
                                     GLib.VariantType('(a(ssssssouso))'),  # reply type
                                     Gio.DBusCallFlags.NONE,
                                     -1,  # timeout
                                     None)  # cancellable

    build_adt_image_object_paths = defaultdict(lambda: defaultdict(dict))
    unit_object_paths = defaultdict(lambda: defaultdict(dict))
    lxd_object_paths = defaultdict(lambda: defaultdict(dict))

    for unit in units:
        (name, _, _, active, _, _, object_path, _, _, _) = unit
        if name.startswith('build-adt-image@') and name.endswith('.timer'):
            name_release_region_arch = name[16:][:-6]
            (release, region, arch) = name_release_region_arch.split('-', -1)
            build_adt_image_object_paths[region][arch][release] = object_path
            continue

        if not name.startswith('autopkgtest@'):
            continue

        name_cloud = name[12:][:-8]

        if name_cloud.startswith('lxd-'):
            # lxd-armhf-1.2.3.4-1
            name_cloud = name_cloud[4:]
            try:
                (arch, ip, n) = name_cloud.split('-', -1)
                lxd_object_paths[arch][ip][n] = object_path
            except ValueError:
                # XXX: for some reason we get autopkgtest@lxd-armhf-a.b.c.d
                # (without the number), not sure why
                pass
            continue

        try:
            (region, arch, n) = name_cloud.split('-', -1)
        except ValueError:
            # autopkgtest@lcy01-1.service
            (region, n) = name_cloud.split('-', -1)
            arch = 'amd64'
        unit_object_paths[region][arch][n] = object_path

    return (unit_object_paths, lxd_object_paths, build_adt_image_object_paths)


def update_cloud_dropins(region, arch, n, releases):
    # re-create, to support fiddling on upgrade

    if arch == 'amd64':
        arches = ['amd64', 'i386']
    else:
        arches = [arch]
    ensure_adt_units = ' '.join(['ensure-adt-image@{}-{}-{}.service'.format(release, region, arch)
                                for release in releases for arch in arches])

    for unit in get_unit_names(region, arch, range(1, n + 1)):
            dropindir = os.path.join(os.path.sep,
                                     'etc',
                                     'systemd',
                                     'system',
                                     '{}.d'.format(unit))
            try:
                os.makedirs(dropindir)
            except FileExistsError:
                shutil.rmtree(dropindir)
                os.makedirs(dropindir)

            with open(os.path.join(dropindir, 'ensure-adt-image.conf'), 'w') as f:
                f.write(dedent('''\
                    [Unit]
                    Requires={ensureadt}
                    After={ensureadt}
                    ''').format(ensureadt=ensure_adt_units))
    reload()


def update_lxd_dropins(arch, ip, n):
    # re-create, to support fiddling on upgrade
    units = ['autopkgtest@lxd-{}-{}-{}.service'.format(arch, ip, m) for m in range(1, n + 1)]

    for unit in units:
        dropindir = os.path.join(os.path.sep,
                                 'etc',
                                 'systemd',
                                 'system',
                                 '{}.d'.format(unit))
        try:
            os.makedirs(dropindir)
        except FileExistsError:
            pass

        with open(os.path.join(dropindir, 'autopkgtest-lxd-socat.conf'), 'w') as f:
            socat_unit = 'autopkgtest-lxd-socat@lxd-{}-{}.service'.format(arch, ip)
            f.write(dedent('''\
                [Unit]
                Requires={socat_unit}
                After={socat_unit}
                ''').format(socat_unit=socat_unit))
    reload()


def enable_timer(region, arch, releases):
    unit_names = ['build-adt-image@{}-{}-{}.timer'.format(release, region, arch) for release in releases]
    enabledisable(unit_names,
                  'EnableUnitFiles',
                  GLib.Variant('(asbb)', (unit_names, False, True)),
                  'StartUnit')
    reload()


def disable_timer(region, arch, releases):
    unit_names = ['build-adt-image@{}-{}-{}.timer'.format(release, region, arch) for release in releases]
    enabledisable(unit_names,
                  'DisableUnitFiles',
                  GLib.Variant('(asb)', (unit_names, False)),
                  'StopUnit')
    reload()


def set_up_systemd_units(target_cloud_config, target_lxd_config, releases):
    """ There are two relevant factors: the units that are currently enabled, and the
        units that the configuration file is asking for. Work out the
        difference between the two, and enable / disable units as required.
        This is split per region-arch. amd64 is a special case. The unit names
        do not have the arch in them. """

    (cloud_unit_object_paths, lxd_unit_object_paths, build_adt_image_object_paths) = get_units()
    all_cloud_regions = set(list(cloud_unit_object_paths.keys()) + list(target_cloud_config.keys()))

    for region in sorted(all_cloud_regions):
        all_arches_for_this_region = set(list(cloud_unit_object_paths[region].keys()) +
                                         list(target_cloud_config.get(region, {}).keys()))
        for arch in sorted(all_arches_for_this_region):
            n_units = len(cloud_unit_object_paths[region][arch])
            target_n_units = target_cloud_config.get(region, {}).get(arch, 0)
            print('Got {target_n_units} units for {region}/{arch}'.format(**locals()))
            if target_n_units > 0:
                update_cloud_dropins(region, arch, target_n_units, releases)
                enable_timer(region, arch, releases)
                if arch == 'amd64':
                    enable_timer(region, 'i386', releases)
            if releases == []:
                target_n_units = 0
            if target_n_units == 0:
                disable_timer(region, arch, releases)
                if arch == 'amd64':
                    disable_timer(region, 'i386', releases)
            if n_units < target_n_units:
                # need to enable some units
                delta = target_n_units - n_units
                unit_names = get_unit_names(region, arch, range(n_units + 1, n_units + delta + 1))
                enable(unit_names)
            elif n_units > target_n_units:
                # need to disable some units
                delta = n_units - target_n_units
                unit_names = get_unit_names(region, arch, range(target_n_units + 1, n_units + 1))
                disable(unit_names)

            currently_enabled_releases = set(build_adt_image_object_paths[region][arch].keys())
            releases_to_disable = currently_enabled_releases - set(releases)

            if releases_to_disable:
                print('Disabling build-adt-image timers for {region}/{arch}/{releases_to_disable}'
                      .format(**locals()))
                disable_timer(region, arch, releases_to_disable)

    # now do lxd. the target config is a dict arch -> IP -> nworkers
    all_lxd_arches = set(list(lxd_unit_object_paths.keys()) + list(target_lxd_config.keys()))
    for arch in sorted(all_lxd_arches):
        all_ips_for_this_arch = set(list(lxd_unit_object_paths[arch].keys()) +
                                    list(target_lxd_config.get(arch, {}).keys()))
        for ip in sorted(all_ips_for_this_arch):
            n_units = len(lxd_unit_object_paths[arch][ip])
            target_n_units = target_lxd_config.get(arch, {}).get(ip, 0)
            if target_n_units > 0:
                update_lxd_dropins(arch, ip, target_n_units)
            if n_units == target_n_units:
                continue
            elif n_units < target_n_units:
                # need to enable some units
                delta = target_n_units - n_units
                unit_names = ['autopkgtest@lxd-{}-{}-{}.service'.format(arch, ip, n)
                              for n in range(n_units + 1, n_units + delta + 1)]
                enable(unit_names)
            else:
                # need to disable some units
                delta = n_units - target_n_units
                unit_names = ['autopkgtest@lxd-{}-{}-{}.service'.format(arch, ip, n)
                              for n in range(target_n_units + 1, n_units + 1)]
                disable(unit_names)


def reload_unit(object_path):
    SYSTEM_BUS.call_sync('org.freedesktop.systemd1',
                         object_path,
                         'org.freedesktop.systemd1.Unit',
                         'Reload',
                         GLib.Variant('(s)', ('replace', )),
                         GLib.VariantType('(o)'),  # reply type
                         Gio.DBusCallFlags.NONE,
                         -1.,  # timeout
                         None)  # cancellable


def reload_units():
    (cloud_unit_object_paths, lxd_unit_object_paths, _) = get_units()

    for region in cloud_unit_object_paths:
        for arch in cloud_unit_object_paths[region]:
            for n in cloud_unit_object_paths[region][arch]:
                object_path = cloud_unit_object_paths[region][arch][n]
                print("Reloading {}".format(object_path))
                reload_unit(object_path)

    for arch in lxd_unit_object_paths:
        for ip in lxd_unit_object_paths[arch]:
            for n in lxd_unit_object_paths[arch][ip]:
                object_path = lxd_unit_object_paths[arch][ip][n]
                print("Reloading {}".format(object_path))
                reload_unit(object_path)
